package com.bergerkiller.bukkit.nolagg.sending;

import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;

import org.bukkit.block.BlockFace;
import org.bukkit.entity.Player;

import com.bergerkiller.bukkit.nolagg.BufferedChunk;
import com.bergerkiller.bukkit.nolagg.Util;

import net.minecraft.server.Chunk;
import net.minecraft.server.EntityPlayer;
import net.minecraft.server.World;

public class ChunkCompressQueue {
    public ChunkCompressQueue(final ChunkSendQueue owner) {
        this.owner = owner;
    }
    private final ChunkSendQueue owner;
    private final LinkedList<ChunkCompressCommand> toCompress = new LinkedList<ChunkCompressCommand>();
    private final LinkedList<ChunkSendCommand> toSend = new LinkedList<ChunkSendCommand>();
    public void update(World world, int x, int z, BlockFace direction) {
    	synchronized (this.toCompress) {
    		for (ChunkCompressCommand cmd : this.toCompress) {
    			cmd.calculateIndex(x, z, direction);
    		}
    		Collections.sort(this.toCompress);
    	}
    	synchronized (this.toSend) {
    		for (ChunkSendCommand cmd : this.toSend) {
    			cmd.calculateIndex(x, z, direction);
    		}
    		Collections.sort(this.toSend);
    	}
    }
    public int getPendingSize() {
    	synchronized (this.toCompress) {
    		synchronized (this.toSend) {
    			return this.toCompress.size() + this.toSend.size();
    		}
    	}
    }
    public void clear() {
    	synchronized (this.toCompress) {
    		this.toCompress.clear();
    	}
    	synchronized (this.toSend) {
    		this.toSend.clear();
    	}
    }
    public boolean isAlive() {
    	if (this.owner.ep.netServerHandler.disconnected) return false;
    	return true;
    }
    public boolean canSend() {
    	synchronized (this.toSend) {
    		return !this.toSend.isEmpty();
    	}
    }
    public boolean finishedCompressing() {
    	synchronized (this.toCompress) {
    		return this.toCompress.isEmpty();
    	}
    }
    public void enqueue(Chunk chunk) {
    	if (chunk == null) return;
        synchronized (this.toCompress) {
            this.toCompress.offer(new ChunkCompressCommand(chunk));
        }
    }
    public void enqueue(ChunkSendCommand sendCommand) {
    	if (sendCommand == null) return;
    	if (!sendCommand.isValid()) return;
        synchronized (this.toSend) {
            this.toSend.offer(sendCommand);
        }
    }
    public boolean remove(int x, int z) {
    	synchronized (this.toCompress) {
    		Iterator<ChunkCompressCommand> iter = this.toCompress.iterator();
    		while (iter.hasNext()) {
    			ChunkCompressCommand cmd = iter.next();
    			if (cmd.chunk.x == x && cmd.chunk.z == z) {
    				iter.remove();
    				return true;
    			}
    		}
    	}
    	synchronized (this.toSend) {
    		Iterator<ChunkSendCommand> iter = this.toSend.iterator();
    		while (iter.hasNext()) {
    			ChunkSendCommand cmd = iter.next();
    			if (cmd.chunk.x == x && cmd.chunk.z == z) {
    				iter.remove();
    				return true;
    			}
    		}
    	}
    	return false;
    }
    public boolean isNear(Chunk chunk) {
    	return this.isNear(chunk.x, chunk.z);
    }
    public boolean isNear(int x, int z) {
    	synchronized (this.owner) {
    		return this.owner.isNear(x, z, Util.view);
    	}
    }
    public EntityPlayer nativeOwner() {
    	return this.owner.ep;
    }
    public Player owner() {
    	return this.owner.ep.getPlayer();
    }
    public Chunk pollChunk() {
    	ChunkCompressCommand chunkcmd;
        synchronized (this.toCompress) {
        	if (this.toCompress.isEmpty()) return null;
        	chunkcmd = this.toCompress.poll();
        }
        Chunk chunk = chunkcmd.chunk;
        if (chunk instanceof BufferedChunk) {
        	if (((BufferedChunk) chunk).isFree()) {
        		synchronized (this.owner) {
        			this.owner.removeContained(chunk.x, chunk.z);
        		}
        		return this.pollChunk();
        	}
        }
        return chunk;
    }
    public ChunkSendCommand pollSendCommand() {
        synchronized (this.toSend) {
        	if (this.toSend.isEmpty()) return null;
        	ChunkSendCommand cmd = this.toSend.poll();
        	if (!this.isNear(cmd.chunk)) {
        		this.owner.removeContained(cmd.chunk.x, cmd.chunk.z);
        		return this.pollSendCommand(); 
        	}
        	return cmd;
        }
    }
    public boolean sendNext() {
    	ChunkSendCommand next = this.pollSendCommand();
        if (next == null) return false;
        next.send(this.owner);
        this.owner.removeContained(next.chunk.x, next.chunk.z);
        return true;
    }
   
}
