package com.bergerkiller.bukkit.nolagg.sending;

import java.util.ArrayList;
import java.util.logging.Level;
import java.util.zip.Deflater;

import lishid.orebfuscator.utils.Calculations;

import com.bergerkiller.bukkit.nolagg.NoLagg;

import net.minecraft.server.Chunk;
import net.minecraft.server.EntityPlayer;

public class ChunkCompressionThread extends Thread {	
	//create a cyclic player compression queue loop
	private static int queueIndex = 0;
	private static ArrayList<ChunkCompressQueue> compress = new ArrayList<ChunkCompressQueue>();
	private static ChunkCompressQueue nextQueue() {
		synchronized (compress) {
			if (compress.isEmpty()) return null;
			if (queueIndex > compress.size() - 1) {
				queueIndex = 0;
			}
			ChunkCompressQueue queue = compress.get(queueIndex);
			if (queue.isAlive()) {
				queueIndex++;
				return queue;
			} else {
				compress.remove(queueIndex);
				return nextQueue();
			}
		}
	}
	public static void addQueue(ChunkCompressQueue queue) {
		synchronized (compress) {
			compress.add(queue);
		}
	}
		
	//thread creation
	private static ChunkCompressionThread[] threads;
	public static void init(int threadcount) {
		if (threadcount <= 0) threadcount = 1;
		if (threads != null) {
			if (threads.length > threadcount) {
				//kill some threads
				ChunkCompressionThread[] newthreads = new ChunkCompressionThread[threadcount];
				for (int i = 0; i < threads.length; i++) {
					if (i < newthreads.length) {
						newthreads[i] = threads[i];
					} else {
						threads[i].interrupt();
					}
				}
				threads = newthreads;
			} else if (threads.length < threadcount) {
				ChunkCompressionThread[] newthreads = new ChunkCompressionThread[threadcount];
				System.arraycopy(threads, 0, newthreads, 0, threads.length);
				for (int i = threads.length; i < newthreads.length; i++) {
					newthreads[i] = new ChunkCompressionThread();
					newthreads[i].start();
				}
				threads = newthreads;
			}
		} else {
			threads = new ChunkCompressionThread[threadcount];
			for (; threadcount > 0; --threadcount) {
				threads[threadcount - 1] = new ChunkCompressionThread();
				threads[threadcount - 1].start();
			}
		}
	}
	public static void deinit() {
		if (threads == null) return;
		for (ChunkCompressionThread thread : threads) {
			thread.interrupt();
		}
		threads = null;
	}
	
	private byte[] rawbuffer = new byte[81920];
	private final byte[] compbuffer = new byte[81920];
    private final Deflater deflater = new Deflater();
    
    private void put(Chunk chunk) {
		System.arraycopy(chunk.b, 0, this.rawbuffer, 0, 32768);
        System.arraycopy(chunk.g.a, 0, this.rawbuffer, 32768, 16384);
        System.arraycopy(chunk.i.a, 0, this.rawbuffer, 49152, 16384);
        System.arraycopy(chunk.h.a, 0, this.rawbuffer, 65536, 16384);
    }
    private void deflate(BufferedPacket51MapChunk packet) {
        this.deflater.reset();
        this.deflater.setLevel(6);
        this.deflater.setInput(packet.rawData);
        this.deflater.finish();
        packet.size = this.deflater.deflate(this.compbuffer);
        if (packet.size == 0) {
        	packet.size = this.deflater.deflate(this.compbuffer);
        }
        packet.prepareBuffer();
        packet.writeBuffer(this.compbuffer);
        packet.rawData = null; //dereference
    }
                                   
    public static double getBusyPercentage(long timescale) {
    	double per = 0;
    	if (threads != null) {
        	for (ChunkCompressionThread thread : threads) {
        		if (thread == null) continue;
        		per += (double) thread.busyDuration / timescale;
        		thread.busyDuration = 0;
        	}
        	per /= threads.length;
    	}
    	return per;
    }
    
    public long busyDuration = 0;
    private long lasttime;

    public BufferedPacket51MapChunk getCompressedPacket(EntityPlayer player, Chunk chunk) {
    	this.lasttime = System.currentTimeMillis();
		//handle a packet
		this.put(chunk);
		//create packet
		BufferedPacket51MapChunk mapchunk = new BufferedPacket51MapChunk(chunk.x, chunk.z, this.rawbuffer);
		//send chunk through possible plugins
		//========================================
		if (NoLagg.isOreObfEnabled) {
			Calculations.Obfuscate(mapchunk, player.getPlayer(), false, false);
		}
		//========================================
		//compress raw data
		this.deflate(mapchunk);
		//actually send this chunk
		this.busyDuration += System.currentTimeMillis() - lasttime;
		return mapchunk;
    }
    
    public final void run() {
    	Chunk chunk;
    	while (!this.isInterrupted()) {
    		try {
    			ChunkCompressQueue queue = nextQueue();
    			if (queue == null || (chunk = queue.pollChunk()) == null) {
    				Thread.sleep(100);
    			} else {
    				try {
    					queue.enqueue(new ChunkSendCommand(this.getCompressedPacket(queue.nativeOwner(), chunk), chunk));
    		    	} catch (Throwable t) {
    		    		NoLagg.log(Level.SEVERE, "Failed to compress map chunk [" + chunk.x + ", " + chunk.z + "] for player " + queue.owner().getName());
    		    		t.printStackTrace();
    		    	}
    			}
    		} catch (InterruptedException ex) {
    		} catch (Throwable t) {
    			t.printStackTrace();
    		}
    	}
    }

}
